@@ -81,13 +81,22 @@ fn builtin_type(ty: &syn::Type) -> u32 {
     }
 }
 
-fn write_u32(val: i32) -> [u8; 4] {
-    [
-        (val & 0xff) as u8,
-        ((val >> 8) & 0xff) as u8,
-        ((val >> 16) & 0xff) as u8,
-        ((val >> 24) & 0xff) as u8,
-    ]
+trait IsVoid {
+    fn is_void(&self) -> bool;
+}
+
+impl IsVoid for syn::Type {
+    fn is_void(&self) -> bool {
+        if let syn::Type::Tuple(tuple) = self {
+            tuple.elems.is_empty()
+        } else {
+            false
+        }
+    }
+}
+
+fn write_i32(vec: &mut Vec<u8>, val: i32) {
+    vec.extend_from_slice(&val.to_le_bytes())
 }
 
 #[derive(Clone)]
@@ -132,6 +141,8 @@ struct MetaObject {
     qt_version: QtVersion,
     int_data: Vec<proc_macro2::TokenStream>,
     meta_types: Vec<proc_macro2::TokenStream>,
+    // Length of string_data vector is guaranteed to be <= i32::MAX.
+    // Each string is guaranteed to be <= i32::MAX too.
     string_data: Vec<String>,
 }
 impl MetaObject {
@@ -146,37 +157,51 @@ impl MetaObject {
 
     fn build_string_data(&self, target_pointer_width: u32) -> Vec<u8> {
         let mut result: Vec<u8> = Vec::new();
+        let r = &mut result;
 
+        // strings are null-terminated, so we push '\0' byte after them and
+        // increment offset couter by an extra 1.
         if self.qt_version == 5 {
-            let sizeof_qbytearraydata = if target_pointer_width == 64 { 24 } else { 16 };
-            let mut ofs = sizeof_qbytearraydata * self.string_data.len() as i32;
+            let sizeof_qbytearraydata: i32 = if target_pointer_width == 64 { 24 } else { 16 };
+            // CAST SAFETY: guaranteed by MetaObject::string_data contract.
+            let mut ofs = sizeof_qbytearraydata.checked_mul(self.string_data.len() as i32).unwrap();
+
             for s in self.string_data.iter() {
-                result.extend_from_slice(&write_u32(-1)); // ref (-1)
-                result.extend_from_slice(&write_u32(s.len() as i32)); // size
-                result.extend_from_slice(&write_u32(0)); // alloc / capacityReserved
+                // CAST SAFETY: guaranteed by MetaObject::string_data contract.
+                let len = s.len() as i32;
+
+                write_i32(r, -1); // ref (-1)
+                write_i32(r, len); // size
+                write_i32(r, 0); // alloc / capacityReserved
                 if target_pointer_width == 64 {
-                    result.extend_from_slice(&write_u32(0)); // padding
+                    write_i32(r, 0); // padding
                 }
-                result.extend_from_slice(&write_u32(ofs)); // offset (LSB)
+                write_i32(r, ofs); // offset (LSB)
                 if target_pointer_width == 64 {
-                    result.extend_from_slice(&write_u32(0)); // offset (MSB)
+                    write_i32(r, 0); // offset (MSB)
                 }
 
-                ofs += s.len() as i32 + 1; // +1 for the '\0'
-                ofs -= sizeof_qbytearraydata;
+                // +1 for the trailing null ('\0')
+                ofs = ofs.checked_add(len).unwrap().checked_add(1).unwrap();
+                ofs = ofs.checked_sub(sizeof_qbytearraydata).unwrap();
             }
         } else {
-            let mut ofs = 2 * 4 * self.string_data.len() as i32;
+            // CAST SAFETY: guaranteed by MetaObject::string_data contract.
+            let mut ofs = (self.string_data.len() as i32).checked_mul(2 * 4).unwrap();
             for s in self.string_data.iter() {
-                result.extend_from_slice(&write_u32(ofs));
-                result.extend_from_slice(&write_u32(s.len() as i32));
-                ofs += s.len() as i32 + 1; // +1 for the '\0'
+                // CAST SAFETY: guaranteed by MetaObject::string_data contract.
+                let len = s.len() as i32;
+
+                write_i32(r, ofs);
+                write_i32(r, len);
+                // +1 for the trailing null ('\0')
+                ofs = ofs.checked_add(len).unwrap().checked_add(1).unwrap();
             }
         }
 
         for s in self.string_data.iter() {
-            result.extend_from_slice(s.as_bytes());
-            result.push(0); // null terminated
+            r.extend_from_slice(s.as_bytes());
+            r.push(0); // null terminator
         }
         result
     }
@@ -340,10 +365,12 @@ impl MetaObject {
     }
 
     fn add_string(&mut self, string: String) -> u32 {
-        if let Some((pos, _)) = self.string_data.iter().enumerate().find(|(_, val)| *val == &string)
-        {
+        if let Some((pos, _)) = self.string_data.iter().enumerate().find(|(_, val)| *val == &string) {
             return pos as u32;
         }
+        assert!(self.string_data.len() < i32::MAX as usize, "String Data: Too many strings registered");
+        assert!(string.len() <= i32::MAX as usize, "String Data: String is too large");
+
         self.string_data.push(string);
         self.string_data.len() as u32 - 1
     }
@@ -461,12 +488,9 @@ pub fn generate(input: TokenStream, is_qobject: bool, qt_version: QtVersion) ->
                             let mut flags = 1 | 2 | 0x00004000 | 0x00001000 | 0x00010000;
                             for it in parsed.1 {
                                 match it {
-                                    Flag::Notify(signal) => {
-                                        assert!(
-                                            notify_signal.is_none(),
-                                            "Two NOTIFY for a property"
-                                        );
-                                        notify_signal = Some(signal);
+                                    Flag::Notify(i) => {
+                                        assert!(notify_signal.is_none(), "Duplicate NOTIFY for a property");
+                                        notify_signal = Some(i);
                                         flags |= 0x00400000;
                                     }
                                     Flag::Const => {
@@ -474,15 +498,15 @@ pub fn generate(input: TokenStream, is_qobject: bool, qt_version: QtVersion) ->
                                         flags &= !2; // Writable
                                     }
                                     Flag::Read(i) => {
-                                        assert!(getter.is_none(), "Two READ for a property");
+                                        assert!(getter.is_none(), "Duplicate READ for a property");
                                         getter = Some(i);
                                     }
                                     Flag::Write(i) => {
-                                        assert!(setter.is_none(), "Two READ for a property");
+                                        assert!(setter.is_none(), "Duplicate READ for a property");
                                         setter = Some(i);
                                     }
                                     Flag::Alias(i) => {
-                                        assert!(alias.is_none(), "Two READ for a property");
+                                        assert!(alias.is_none(), "Duplicate READ for a property");
                                         alias = Some(i);
                                     }
                                 }
@@ -633,6 +657,7 @@ pub fn generate(input: TokenStream, is_qobject: bool, qt_version: QtVersion) ->
         quote! {
             let pinned = <#name #ty_generics as #crate_::QObject>::get_from_cpp(o);
             // FIXME: we should probably use borrow_mut here instead, but in a way which order re-entry
+            #[allow(unused_variables)]
             let mut obj = &mut *pinned.as_ptr();
 
             assert_eq!(o, obj.get_cpp_object(), "Internal pointer invalid");
@@ -653,6 +678,7 @@ pub fn generate(input: TokenStream, is_qobject: bool, qt_version: QtVersion) ->
         }
     } else {
         quote! {
+            #[allow(unused_variables)]
             let mut obj = ::std::mem::transmute::<*mut ::std::os::raw::c_void, &mut #name #ty_generics>(o);
         }
     };
@@ -679,20 +705,6 @@ pub fn generate(input: TokenStream, is_qobject: bool, qt_version: QtVersion) ->
                 };
             }
 
-            let register_type = if builtin_type(&prop.typ) == 0 {
-                let typ_str = typ.clone().into_token_stream().to_string();
-                let typ_str = typ_str.as_bytes();
-                quote! {
-                    #RegisterPropertyMetaType => unsafe {
-                        let r = *a as *mut i32;
-                        *r = <#typ as #crate_::PropertyType>::register_type(
-                            ::std::ffi::CStr::from_bytes_with_nul_unchecked(&[#(#typ_str ,)* 0u8]) );
-                    }
-                }
-            } else {
-                quote!{}
-            };
-
             let getter = if let Some(ref getter) = prop.getter {
                 let getter_ident: syn::Ident = getter.clone();
                 quote!{
@@ -715,19 +727,40 @@ pub fn generate(input: TokenStream, is_qobject: bool, qt_version: QtVersion) ->
                 }
             };
 
-            quote! { #i => match c {
-                #ReadProperty => unsafe {
-                    #get_object
-                    #getter
-                },
-                #WriteProperty => unsafe {
-                    #get_object
-                    #setter
-                },
-                #ResetProperty => { /* TODO */},
-                #register_type
-                _ => {}
-            }}
+            // register properties of non-built-in types: stringify type's TokenStream
+            let register_type = if builtin_type(typ) == 0 {
+                let typ_str = typ.clone().into_token_stream().to_string();
+                let typ_bytes = typ_str.as_bytes();
+                quote! { /* externally defined variables: register_result. */
+                    // SAFETY: string generated from Rust type tokens should be a valid C string.
+                    let name = unsafe { ::std::ffi::CStr::from_bytes_with_nul_unchecked(&[#(#typ_bytes ,)* 0u8]) };
+                    *register_result = <#typ as #crate_::PropertyType>::register_type(name);
+                }
+            } else {
+                quote! {}
+            };
+
+            quote! {
+                #i => match c {
+                    #ReadProperty => unsafe {
+                        #get_object
+                        #getter
+                    },
+                    #WriteProperty => unsafe {
+                        #get_object
+                        #setter
+                    },
+                    #ResetProperty => {/* TODO */},
+                    #RegisterPropertyMetaType => {
+                         // a[0]: registerResult, should set to id of registered type or -1 if a type
+                         // could not be registered for any reason.
+                         // SAFETY: Qt always passes a valid pointer here.
+                        let register_result: &mut i32 = unsafe { &mut *((*a.offset(0)) as *mut i32) };
+                        #register_type
+                    },
+                    _ => {}
+                }
+            }
         })
         .collect();
 
@@ -745,28 +778,30 @@ pub fn generate(input: TokenStream, is_qobject: bool, qt_version: QtVersion) ->
                     let i = i as isize;
                     let ty = &arg.typ;
                     quote! {
+                        // a[1..=N] are pointers to the arguments
+                        // References to the builtin types are reinterpreted as their Rust
+                        // counterparts using `builtin_type()` mapping.
                         (*(*(a.offset(#i + 1)) as *const #ty)).clone()
                     }
                 })
                 .collect();
 
-            fn is_void(ret_type: &syn::Type) -> bool {
-                if let syn::Type::Tuple(ref tuple) = ret_type {
-                    tuple.elems.is_empty()
-                } else {
-                    false
-                }
-            }
+            let call = quote! { obj.#method_name(#(#args_call),*) };
 
-            if is_void(&method.ret_type) {
-                quote! { #i => obj.#method_name(#(#args_call),*), }
+            if method.ret_type.is_void() {
+                quote! { #i => #call, }
             } else {
                 let ret_type = &method.ret_type;
-                let args_call2 = args_call.clone();
-                quote! { #i => {
-                        let r = *a as *mut #ret_type;
-                        if r.is_null() { obj.#method_name(#(#args_call),*); }
-                        else { *r = obj.#method_name(#(#args_call2),*); }
+                quote! {
+                    #i => {
+                        let r = #call;
+                        // a[0] is a pointer to the return value
+                        // SAFETY: pointer to the return value is guaranteed to exist in array `a`,
+                        // but it may be null if the caller is not interested in reading it.
+                        let return_value = unsafe { (*a.offset(0)) as *mut #ret_type };
+                        if let Some(return_ref) = unsafe { return_value.as_mut() } {
+                            *return_ref = r;
+                        }
                     }
                 }
             }
@@ -776,34 +811,41 @@ pub fn generate(input: TokenStream, is_qobject: bool, qt_version: QtVersion) ->
     let register_arguments: Vec<_> = methods
         .iter()
         .enumerate()
-        .map(|(i, method)| {
-            let i = i as u32;
+        .map(|(fn_i, method)| {
+            let fn_i = fn_i as u32;
             let args: Vec<_> = method
                 .args
                 .iter()
                 .enumerate()
-                .map(|(i, arg)| {
-                    let i = i as u32;
-                    if builtin_type(&arg.typ) == 0 {
-                        let typ = &arg.typ;
-                        let typ_str = arg.typ.clone().into_token_stream().to_string();
+                .map(|(arg_i, arg)| {
+                    let arg_i = arg_i as u32;
+                    let typ = &arg.typ;
+                    // FIXME: there's minor code duplication with `property_meta_call`
+                    if builtin_type(&typ) == 0 {
+                        let typ_str = typ.clone().into_token_stream().to_string();
                         let typ_str = typ_str.as_bytes();
-                        quote! {
-                            #i => { unsafe { *(*a as *mut i32) = <#typ as #crate_::QMetaType>::register(
-                                Some(::std::ffi::CStr::from_bytes_with_nul_unchecked(&[#(#typ_str ,)* 0u8]))) }; }
+                        quote! { /* externally defined variables: arg_type, arg_index. */
+                            #arg_i => {
+                                // SAFETY: string generated from Rust type tokens should be a valid C string.
+                                let name = unsafe { ::std::ffi::CStr::from_bytes_with_nul_unchecked(&[#(#typ_str ,)* 0u8]) };
+                                let ty = <#typ as #crate_::QMetaType>::register(Some(name));
+                                ty
+                            }
                         }
                     } else {
-                        quote!{}
+                        quote! {}
                     }
                 })
                 .collect();
 
-            quote! { #i => {
-                match unsafe { *(*(a.offset(1)) as *const u32) } {
-                    #(#args)*
-                    _ => {}
+            quote! { /* externally defined variables: arg_type, arg_index. */
+                #fn_i => {
+                    match arg_index {
+                        #(#args)*
+                        _ => -1, // default when type is unknown
+                    }
                 }
-            }}
+            }
         })
         .collect();
 
@@ -846,17 +888,21 @@ pub fn generate(input: TokenStream, is_qobject: bool, qt_version: QtVersion) ->
         }
     }));
 
+    // Despite its name, it actually handles signals.
     let index_of_method = signals.iter().enumerate().map(|(i, signal)| {
         let sig_name = &signal.name;
-        // if *a[1] == offset_of(signal field)  =>  *a[0] = index and return.
-        quote! {
-            unsafe {
-                let null = ::std::ptr::null() as *const #name #ty_generics;
-                let offset = &(*null).#sig_name as *const _ as usize - (null as usize);
-                if (*(*(a.offset(1)) as *const usize)) == offset  {
-                    *(*a as *mut i32) = #i as i32;
-                    return;
-                }
+        // if signal == offset_of(signal field) then *result = index and return.
+        quote! { /* externally defined variables: signal, result. */
+            // SAFETY: no dereference of null pointer, only calculation of field's offset over
+            // imaginary struct located at null. In Rust null is 0, thus aligned for any type.
+            let offset = unsafe {
+                let base = ::std::ptr::null() as *const #name #ty_generics;
+                let field = (&(*base).#sig_name) as *const _ as usize;
+                field - (base as usize)
+            };
+            if signal == offset  {
+                *result = #i as i32;
+                return;
             }
         }
     });
@@ -869,16 +915,18 @@ pub fn generate(input: TokenStream, is_qobject: bool, qt_version: QtVersion) ->
 
     let mo = if ast.generics.params.is_empty() {
         quote! {
-            #crate_::qmetaobject_lazy_static! { static ref MO: #crate_::QMetaObject = #crate_::QMetaObject {
-                super_data: #base_meta_object,
-                #super_data_getter
-                string_data: STRING_DATA.as_ptr(),
-                data: INT_DATA.as_ptr(),
-                static_metacall: Some(static_metacall),
-                related_meta_objects: ::std::ptr::null(),
-                meta_types: #meta_types_init,
-                extra_data: ::std::ptr::null(),
-            };};
+            #crate_::qmetaobject_lazy_static! {
+                static ref MO: #crate_::QMetaObject = #crate_::QMetaObject {
+                    super_data: #base_meta_object,
+                    #super_data_getter
+                    string_data: STRING_DATA.as_ptr(),
+                    data: INT_DATA.as_ptr(),
+                    static_metacall: Some(static_metacall),
+                    related_meta_objects: ::std::ptr::null(),
+                    meta_types: #meta_types_init,
+                    extra_data: ::std::ptr::null(),
+                };
+            };
             return &*MO;
         }
     } else {
@@ -916,7 +964,7 @@ pub fn generate(input: TokenStream, is_qobject: bool, qt_version: QtVersion) ->
 
     let qobject_spec_func = if is_qobject {
         quote! {
-            fn get_cpp_object(&self)-> *mut ::std::os::raw::c_void {
+            fn get_cpp_object(&self) -> *mut ::std::os::raw::c_void {
                 self.#base_prop.get()
             }
 
@@ -1004,18 +1052,36 @@ pub fn generate(input: TokenStream, is_qobject: bool, qt_version: QtVersion) ->
                     a: *const *mut ::std::os::raw::c_void
                 ) #where_clause
                 {
-                    if c == #InvokeMetaMethod { unsafe {
-                        #get_object
-                        match idx {
-                            #(#method_meta_call)*
-                            _ => { let _ = obj; }
+                    if c == #InvokeMetaMethod {
+                        // a[0]: pointer to return value.
+                        // a[1..=10]: pointers to arg0..arg9.
+                        unsafe {
+                            #get_object
+                            match idx {
+                                #(#method_meta_call)*
+                                _ => {}
+                            }
                         }
-                    }} else if c == #IndexOfMethod - #qt6_offset {
+                    } else if c == #IndexOfMethod - #qt6_offset {
+                        // a[0]: pointer to an index of the method, return value.
+                        let result: &mut i32 = unsafe { &mut *((*a.offset(0)) as *mut i32) };
+                        // a[1]: pointer to raw signal representation (in our case it's field's offset).
+                        // SAFETY: Qt never passes us signal representations of other classes, so
+                        // it is guaranteed that this representation was generated by our code,
+                        // thus known to be safe.
+                        let signal: usize = unsafe { *((*a.offset(1)) as *const usize) };
                         #(#index_of_method)*
                     } else if c == #RegisterMethodArgumentMetaType - #qt6_offset {
-                        match idx {
+                        // a[0]: pointer to type of the argument at given index, return value.
+                        // SAFETY: Qt always passes a valid pointer here, but it is 'out' parameter,
+                        // thus may not be initialized and must not be read from.
+                        let arg_type: &mut i32 = unsafe { &mut *((*a.offset(0)) as *mut i32) };
+                        // a[1]: pointer to index of the requested argument.
+                        // SAFETY: Qt always passes a valid pointer here and checks that index is >= 0.
+                        let arg_index: u32 = unsafe { *((*a.offset(1)) as *const u32) };
+                        *arg_type = match idx {
                             #(#register_arguments)*
-                            _ => {}
+                            _ => -1, // default when type is unknown
                         }
                     } else {
                         match idx {
@@ -1032,7 +1098,6 @@ pub fn generate(input: TokenStream, is_qobject: bool, qt_version: QtVersion) ->
         }
 
     };
-
     if is_plugin {
         use crate::qbjs::Value;
         let mut object_data: Vec<(&'static str, Value)> = vec![
@@ -1046,7 +1111,10 @@ pub fn generate(input: TokenStream, is_qobject: bool, qt_version: QtVersion) ->
 
         let plugin_data = qbjs::serialize(&object_data);
         let plugin_data_size = plugin_data.len();
-        body = quote! { #body
+
+        body = quote! {
+            #body
+
             #[cfg_attr(target_os = "macos", link_section = "__TEXT,qtmetadata")]
             #[cfg_attr(not(target_os = "macos"), link_section = ".qtmetadata")]
             #[no_mangle]
@@ -1078,8 +1146,8 @@ fn is_valid_repr_attribute(attribute: &syn::Attribute) -> bool {
             if list.path.is_ident("repr") && list.nested.len() == 1 {
                 match &list.nested[0] {
                     syn::NestedMeta::Meta(syn::Meta::Path(word)) => {
-                        const ACCEPTABLES: &[&str] = &["u8", "u16", "u32", "i8", "i16", "i32"];
-                        ACCEPTABLES.iter().any(|w| word.is_ident(w))
+                        const ACCEPTABLE_TYPES: &[&str] = &["u8", "u16", "u32", "i8", "i16", "i32"];
+                        ACCEPTABLE_TYPES.iter().any(|w| word.is_ident(w))
                     }
                     _ => false,
                 }
@@ -1156,16 +1224,19 @@ pub fn generate_enum(input: TokenStream, qt_version: QtVersion) -> TokenStream {
 
     let mo = if ast.generics.params.is_empty() {
         quote! {
-            #crate_::qmetaobject_lazy_static! { static ref MO: #crate_::QMetaObject = #crate_::QMetaObject {
-                super_data: ::std::ptr::null(),
-                #super_data_getter
-                string_data: STRING_DATA.as_ptr(),
-                data: INT_DATA.as_ptr(),
-                static_metacall: None,
-                related_meta_objects: ::std::ptr::null(),
-                meta_types: ::std::ptr::null(),
-                extra_data: ::std::ptr::null(),
-            };};
+            #crate_::qmetaobject_lazy_static! {
+                static ref MO: #crate_::QMetaObject = #crate_::QMetaObject {
+                    super_data: ::std::ptr::null(),
+                    #super_data_getter
+                    string_data: STRING_DATA.as_ptr(),
+                    data: INT_DATA.as_ptr(),
+                    static_metacall: None,
+                    related_meta_objects: ::std::ptr::null(),
+                    meta_types: ::std::ptr::null(),
+                    extra_data: ::std::ptr::null(),
+                };
+            };
+
             return &*MO;
         }
     } else {
@@ -1176,7 +1247,7 @@ pub fn generate_enum(input: TokenStream, qt_version: QtVersion) -> TokenStream {
         impl #crate_::QEnum for #name {
             fn static_meta_object() -> *const #crate_::QMetaObject {
                 #str_data
-                static INT_DATA : &'static [u32] = & [ #(#int_data),* ];
+                static INT_DATA : &'static [u32] = &[ #(#int_data),* ];
 
                 #mo
             }
